
import 'server.dart';

@ComposeSubtypes
abstract class StatusActionCheck {

  @Inject
  Server get server;

  Future<Map<String,bool>> check();
}

abstract class StatusActionDbCheck extends StatusActionCheck {
  Future<Map<String,bool>> check() async {
    return {'database' : await server.db.fetchOne('SELECT 1') == 1};
  }
}

abstract class StatusActionTickerCheck extends StatusActionCheck {
  Future<Map<String,bool>> check() async {
    int serviceId = server.config.getRequired<int>('service_id');
    var lastRun = await server.db.fetchOne(
        'SELECT last_run FROM run_jobs WHERE app_id = ? AND job = ?',
        [ serviceId, 'Ticker' ]
    );
    var ret = false;
    if (lastRun != null) {
      DateTime daemonLastRun = server.db.fixTZ(lastRun);
      ret = new DateTime.now().difference(daemonLastRun).inSeconds < 65;
    }
    return {'ticker' : ret};
  }
}

/**
 * Status Action
 */
abstract class StatusAction extends JsonAction {


  @InjectInstances
  Map<String, StatusActionCheck> get allChecks;

  Future<dynamic> run() async
  {
    Map<String, bool> checks = {};
    for (var key in allChecks.keys) {
      try {
        checks.addAll(await allChecks[key]!.check());
      } catch (e) {
        checks[key] = false;
      }
    }

    var healthy = checks.values.reduce((a, b) => a && b);
    if (!healthy) {
      this.responseStatus = HttpStatus.serviceUnavailable;
    }
    return {
      'time': DateTime.now().toString(),
      'healthy': healthy,
      'checks': checks
    };
  }

}

/**
 * Schema Action
 */
abstract class SchemaAction extends JsonAction {

  Map<String, dynamic> routeNodeToMap(String path, RouteNode node)
  {
    Map<String, dynamic> ret = {};
    if (node.className != null) {
      ret[path] = {
        'handler': node.className,
        'params': {}
      };
    }
    node.subNodes.forEach((k, node){
      ret.addAll(routeNodeToMap(path + '/' + k, node));
    });
    return ret;
  }

  Future<dynamic> run() async
  {
    return routeNodeToMap('', server.routing.root);
  }

}

/**
 * Robots Action
 */
abstract class RobotsAction extends HttpAction {

  String? ext = 'txt';

  Future handleRequest() async
  {
    request.response.statusCode = 200;
    request.response.headers.contentType = ContentType.text;
    request.response.writeln('User-agent: *');
    request.response.writeln('Disallow: /');
  }

}


/**
 * Favicon Action
 */
abstract class FaviconAction extends HttpAction {

  String? ext = 'ico';

  Future handleRequest() async
  {
    request.response.statusCode = 200;
    request.response.headers.contentType = new ContentType('image', 'svg+xml');
    request.response.write('<?xml version="1.0" encoding="UTF-8" standalone="no"?><svg xmlns="http://www.w3.org/2000/svg" height="100" width="100"><path d="M 60,100 C 42,100 20,94 9,79 39,97 80,85 84,60 89,35 71,23 55,22 33,22 9,35 31,59 35,65 51,72 59,59 73,35 25,35 50,56 33,57 30,45 35,39 38,34 56,30 65,36 82,50 64,74 48,79 29,85 9,72 3,55 -10,16 25,0 55,0 85,0 100,29 100,52 100,76 80,99 60,100 Z" style="fill:#4582b2"/></svg>');
  }

}
